@@ -158,6 +158,7 @@
 
 #include <msg.h>
 #include <msg_syslog.h>
+#include <msg_vstream.h>
 #include <chroot_uid.h>
 #include <vstring.h>
 #include <vstream.h>
@@ -261,7 +262,7 @@ static void trigger_server_wakeup(int fd)
 
 static void trigger_server_accept_fifo(int unused_event, char *context)
 {
-    char   *myname = "trigger_server_accept_fifo";
+    const char *myname = "trigger_server_accept_fifo";
     int     listen_fd = CAST_CHAR_PTR_TO_INT(context);
 
     if (trigger_server_lock != 0
@@ -285,7 +286,7 @@ static void trigger_server_accept_fifo(int unused_event, char *context)
 
 static void trigger_server_accept_local(int unused_event, char *context)
 {
-    char   *myname = "trigger_server_accept_local";
+    const char *myname = "trigger_server_accept_local";
     int     listen_fd = CAST_CHAR_PTR_TO_INT(context);
     int     time_left = 0;
     int     fd;
@@ -331,7 +332,7 @@ static void trigger_server_accept_local(int unused_event, char *context)
 
 static void trigger_server_accept_pass(int unused_event, char *context)
 {
-    char   *myname = "trigger_server_accept_pass";
+    const char *myname = "trigger_server_accept_pass";
     int     listen_fd = CAST_CHAR_PTR_TO_INT(context);
     int     time_left = 0;
     int     fd;
@@ -377,7 +378,7 @@ static void trigger_server_accept_pass(int unused_event, char *context)
 
 NORETURN trigger_server_main(int argc, char **argv, TRIGGER_SERVER_FN service,...)
 {
-    char   *myname = "trigger_server_main";
+    const char *myname = "trigger_server_main";
     char   *root_dir = 0;
     char   *user_name = 0;
     int     debug_me = 0;
@@ -403,6 +404,7 @@ NORETURN trigger_server_main(int argc, char **argv, TRIGGER_SERVER_FN service,..
     WATCHDOG *watchdog;
     char   *oval;
     char   *generation;
+    int     msg_vstream_needed = 0;
 
     /*
      * Process environment options as early as we can.
@@ -454,7 +456,7 @@ NORETURN trigger_server_main(int argc, char **argv, TRIGGER_SERVER_FN service,..
      * stderr, because no-one is going to see them.
      */
     opterr = 0;
-    while ((c = GETOPT(argc, argv, "cdDi:lm:n:o:s:St:uvz")) > 0) {
+    while ((c = GETOPT(argc, argv, "cdDi:lm:n:o:s:St:uvVz")) > 0) {
 	switch (c) {
 	case 'c':
 	    root_dir = "setme";
@@ -498,6 +500,10 @@ NORETURN trigger_server_main(int argc, char **argv, TRIGGER_SERVER_FN service,..
 	case 'v':
 	    msg_verbose++;
 	    break;
+	case 'V':
+	    if (++msg_vstream_needed == 1)
+		msg_vstream_init(mail_task(var_procname), VSTREAM_ERR);
+	    break;
 	case 'z':
 	    zerolimit = 1;
 	    break;
@@ -672,9 +678,7 @@ NORETURN trigger_server_main(int argc, char **argv, TRIGGER_SERVER_FN service,..
      * Optionally, restrict the damage that this process can do.
      */
     resolve_local_init();
-#ifdef SNAPSHOT
     tzset();
-#endif
     chroot_uid(root_dir, user_name);
 
     /*
